"""
These are global fixtures automagically loaded by pytest
"""

import pytest
from srgutil.context import default_context
from unittest.mock import Mock
from threading import Lock


@pytest.fixture
def MOCK_DATA():
    return {"guid-1": {'guid-2': 1000,
                       'guid-3': 100,
                       'guid-4': 10,
                       'guid-5': 1,
                       'guid-6': 1},
            'guid-2': {'guid-1': 50,
                       'guid-3': 40,
                       'guid-4': 20,
                       'guid-8': 30,
                       'guid-9': 10},
            'guid-3': {'guid-1': 100,
                       'guid-2': 40,
                       'guid-4': 70},
            'guid-4': {'guid-2': 20},
            "guid-6": {'guid-1': 5,
                       'guid-7': 100,
                       'guid-8': 100,
                       'guid-9': 100},
            'guid-8': {'guid-2': 30},
            'guid-9': {'guid-2': 10},
            }


@pytest.fixture
def TIE_MOCK_DATA():
    return {"guid-1": {'guid-2': 100,
                       'guid-3': 100,
                       'guid-4': 100,
                       'guid-5': 100},
            'guid-2': {'guid-1': 100,
                       'guid-3': 100,
                       'guid-4': 100,
                       'guid-5': 100},
            'guid-3': {'guid-1': 100,
                       'guid-2': 100,
                       'guid-4': 100,
                       'guid-5': 100},
            'guid-4': {'guid-1': 20,
                       'guid-2': 20,
                       'guid-3': 20,
                       'guid-5': 20},
            'guid-5': {'guid-1': 20,
                       'guid-2': 20,
                       'guid-3': 20,
                       'guid-4': 20},
            }


@pytest.fixture
def MOCK_GUID_RANKING():
    return {"guid-1": 10,
            'guid-2': 9,
            'guid-3': 8,
            'guid-4': 7,
            'guid-5': 6,
            "guid-6": 5,
            'guid-7': 4,
            'guid-8': 3,
            'guid-9': 2}


@pytest.fixture
def CUTOFF_GUID_RANKING():
    return {"guid-1": 10000,
            'guid-2': 9000,
            'guid-3': 8000,
            'guid-4': 7,
            'guid-5': 6000,
            "guid-6": 5000,
            'guid-7': 4000,
            'guid-8': 3000,
            'guid-9': 2000}


@pytest.fixture
def default_ctx():
    """
    This sets up a basic context for use for testing
    """
    ctx = default_context()
    ctx['ignore_redis'] = True
    ctx['CACHE_URL'] = "redis://redis:6369/0"
    return ctx


def mock_cold_redis_cache(json_loader):
    """
    Mock out the methods of the jsonloader and return it
    """
    json_loader._redis = Mock()

    # Set the redis cache as cold
    json_loader._redis.exists.return_value = False

    # Set the redis cache as cold
    json_loader._redis.get.return_value = None

    json_loader._redis.lock.return_value = Lock()

    # Set the locally cached JSON data object to None
    json_loader._cached_copy = None

    return json_loader



from moto import mock_s3
from taar_lite.recommenders import GuidBasedRecommender  # noqa
from taar_lite.recommenders.cache import LazyJSONLoader  # noqa
from taar_lite.recommenders.guid_based_recommender import ADDON_LIST_BUCKET
from taar_lite.recommenders.guid_based_recommender import ADDON_LIST_KEY
from taar_lite.recommenders.guid_based_recommender import GUID_RANKING_KEY
import boto3
import json
from .conftest import mock_cold_redis_cache
from unittest.mock import Mock

from contextlib import contextmanager


def install_mock_data(MOCK_DATA, MOCK_GUID_RANKING):
    conn = boto3.resource('s3', region_name='us-west-2')

    conn.create_bucket(Bucket=ADDON_LIST_BUCKET)

    conn.Object(ADDON_LIST_BUCKET, ADDON_LIST_KEY)\
        .put(Body=json.dumps(MOCK_DATA))
    conn.Object(ADDON_LIST_BUCKET, GUID_RANKING_KEY)\
        .put(Body=json.dumps(MOCK_GUID_RANKING))


def test_get_json_hot_cache(default_ctx):
    coinstall_loader = LazyJSONLoader(default_ctx,
                                      ADDON_LIST_BUCKET,
                                      ADDON_LIST_KEY)

    coinstall_loader._redis = Mock()

    # Mock out the lock
    @contextmanager
    def MockContextManager(*args, **kwargs):
        pass
        yield
        pass

    coinstall_loader._redis.lock.return_value = MockContextManager()

    # Set the redis cache as hot
    coinstall_loader._redis.exists.return_value = True
    # Set the locally cached JSON data object
    coinstall_loader._cached_copy = {"payload": "hot cached copy"}

    actual = coinstall_loader.get()
    assert actual == {"payload": "hot cached copy"}


@mock_s3
def test_get_json_cold_cache(default_ctx, MOCK_DATA, MOCK_GUID_RANKING):
    install_mock_data(MOCK_DATA, MOCK_GUID_RANKING)
    coinstall_loader = LazyJSONLoader(default_ctx,
                                      ADDON_LIST_BUCKET,
                                      ADDON_LIST_KEY)

    mock_cold_redis_cache(coinstall_loader)

    actual = coinstall_loader.get()

    # The data from the S3 mocking should be loaded here
    assert actual == MOCK_DATA



import redis
import boto3
import json
from srgutil.interfaces import IMozLogging


class LazyJSONLoader:
    def __init__(self, ctx, s3_bucket, s3_key, expiry=14400):
        self._ctx = ctx

        self._redis = redis.StrictRedis.from_url(ctx['CACHE_URL'])

        self.logger = self._ctx[IMozLogging].get_logger('srgutil')

        self._s3_bucket = s3_bucket
        self._s3_key = s3_key

        self._key_str = "{}|{}".format(self._s3_bucket, self._s3_key)

        self.logger = self._ctx[IMozLogging].get_logger('srgutil')

        self._expiry_seconds = expiry
        self.logger.info("Set expiry time to {}".format(self._expiry_seconds))

        self._cached_copy = None

    def force_expiry(self):
        key_str = "{}|{}".format(self._s3_bucket, self._s3_key)

        # Clear the byte cache in redis as well as locally cached copy
        self._redis.delete(key_str)
        self._cached_copy = None

    def get(self):
        """Fetch the value that is stored"""

        raw_data = None
        raw_bytes = None
        try:
            # We must acquire a lock to redis as we're checking across
            # the redis cache and the local _cached_copy of data.
            # The redis cache is a distributed lock so we don't need
            # to lock the process using python thread locks.
            #
            # Note that we need to hold the redis lock until we
            # return a value or have updated redis with new data
            # and a new expiration time.
            with self._redis.lock("lock|{}".format(self._key_str), timeout=60):
                # If the redis cache is hot and we have a cached copy
                # still, just return the cached copy
                if self._redis.exists(self._key_str) and self._cached_copy is not None:
                    return self._cached_copy

            # Loading cache data from redis doesn't need to be in a
            # lock.  The data will either be there or it won't.
            #
            # If the data is available in redis, we update the process
            # local copy of _cached_copy.
            #
            # If the data is not available in redis, we load from S3,
            # write the bytes to redis and then update the local copy
            # of _cached_copy.
            #
            # Neither case needs to be run in a lock as the worst case
            # scenario is that multiple threads will concurrently
            # write the same data to redis and the same data will
            # be decoded on each execution thread and stored in
            # _cached_copy.
            raw_bytes = self._redis.get(self._key_str)

            if raw_bytes is None:
                # The raw_bytes data has expired from the redis cache.
                # We need to force a data reload from S3
                s3 = boto3.resource('s3')
                raw_bytes = (
                    s3
                    .Object(self._s3_bucket, self._s3_key)
                    .get()['Body']
                    .read()
                )
                msg = "Loaded JSON from S3: {}".format(self._key_str)
                self.logger.info(msg)
                self._redis.set(self._key_str, raw_bytes, ex=self._expiry_seconds)
            else:
                # redis has a cached copy of the dataset.  Just
                # continue on to decode the data
                pass

            raw_data = (
                raw_bytes.decode('utf-8')
            )

            # It is possible to have corrupted files in S3, so
            # protect against that.
            try:
                self._cached_copy = json.loads(raw_data)
            except ValueError:
                # Explicitly set the local cached_copy to None
                self._cached_copy = None

                self.logger.error("Cannot parse JSON resource from S3", extra={
                    "bucket": self._s3_bucket,
                    "key": self._s3_key})

            return self._cached_copy
        except Exception:
            self.logger.exception("Failed to download from S3", extra={
                "bucket": self._s3_bucket,
                "key": self._s3_key})
            return None
